# -*- coding: utf-8 -*-

#  PycaWM
#  focuses.py
#  Copyright (c) 2007 Vincent Rasneur, Anaël Verrier

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 3 only.

#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys

if sys.version_info < (2, 5):
    raise ImportError('focuses module needs at least Python 2.5!')

from Xlib import X

from decorators import managed_client
from hookmanager import (add_overriding_hook, remove_overriding_hook,
                         add_pre_hook, remove_pre_hook,
                         add_post_hook, remove_post_hook)

from client import Client

# Note for the focuses: All the methods which have the same name has the wm
# ones will be automatically added as post hooks

class Focus(object):
    def init_focus(self, wm):
        pass

    def destroy_focus(self, wm):
        pass

class ClickToFocus(Focus):
    def init_focus(self, wm):
        add_post_hook(Client.take_focus, Client.upper)
        add_pre_hook(wm.button_press_reply, self.pre_button_press_reply)
        self.init_focus_aux(wm)

    def init_focus_aux(self, wm):
        # this method will be reused in ClickToSloppyFocus
        add_post_hook(Client.__init__, self.client_grab_button)
        if wm.clients:
            client = wm.clients[0]
            if not client.is_hidden():
                client.take_focus()
        for client in wm.clients:
            client.grab_button()

    @staticmethod
    def client_grab_button(client, *args, **kwargs):
        client.grab_button()

    @staticmethod
    @managed_client
    def pre_button_press_reply(wm, event, client):
        # pre hook, if we put a post hook the information might be bad
        # when we will have 'actions' (the button action may have been done)
        if event.window == client.window:
            wm.display.allow_events(X.ReplayPointer, X.CurrentTime)
        client.take_focus()

    @staticmethod
    @managed_client
    def map_request_reply(wm, event, client):
        # a mapped window (in a visible state) should always have the focus
        # and be the first in the clients stack
        if not client.is_hidden():
            client.take_focus()

    def destroy_focus_aux(self, wm):
        # this method will be reused in ClickToSloppyFocus
        remove_pre_hook(wm.button_press_reply, self.pre_button_press_reply)
        remove_post_hook(Client.__init__, self.client_grab_button)
        # a bit harsh, but it works...
        for client in wm.clients:
            client.ungrab_button()

    def destroy_focus(self, wm):
        remove_post_hook(Client.take_focus, Client.upper)
        self.destroy_focus_aux(wm)

class SloppyFocus(Focus):
    def init_focus(self, wm):
        add_pre_hook(wm.button_press_reply, self.pre_button_press_reply)
        if not (wm.clients and wm.clients[0].states.focused):
            SloppyFocus.simulate_enter_reply(wm)

    @staticmethod
    def simulate_enter_reply(wm):
        pointer = wm.root.query_pointer()
        client = wm.client_at_point(pointer.root_x, pointer.root_y)
        if client is not None:
            client.take_focus()
        else:
            wm.display.set_input_focus(X.PointerRoot, X.RevertToNone,
                                       X.CurrentTime)
            if wm.clients and wm.clients[0].states.focused:
                wm.clients[0].leave_focus()
            wm.display.sync()

    @staticmethod
    @managed_client
    def pre_button_press_reply(wm, event, client):
        # pre hook too, for the same reason as ClickToFocus
        if (event.window == client.title_bar and
            client.button_pressed(event.event_x, event.event_y) is None):
            if event.detail == X.Button1:
                client.upper()
            elif event.detail == X.Button3:
                client.lower()

    @staticmethod
    @managed_client
    def enter_reply(wm, event, client):
        client.take_focus()

    def destroy_focus(self, wm):
        remove_pre_hook(wm.button_press_reply, self.pre_button_press_reply)

class FocusFollowsMouse(SloppyFocus):
    def init_focus(self, wm):
        add_overriding_hook(wm.give_focus, FocusFollowsMouse.give_focus)
        add_pre_hook(wm.button_press_reply, SloppyFocus.pre_button_press_reply)
        SloppyFocus.simulate_enter_reply(wm)

    @staticmethod
    @managed_client
    def leave_reply(wm, event, client):
        if (event.detail == X.NotifyVirtual or
            (event.window == client.parent and
             event.detail == X.NotifyAncestor)):
            client.leave_focus()
            wm.root.set_input_focus(X.RevertToPointerRoot, X.CurrentTime)

    @staticmethod
    def give_focus(wm, client):
        pointer = wm.root.query_pointer()
        if wm.client_has_pointer(pointer.root_x, pointer.root_y):
            wm.give_focus._wrapped_fun(wm, client)

    def destroy_focus(self, wm):
        remove_overriding_hook(wm.give_focus, FocusFollowsMouse.give_focus)
        remove_pre_hook(wm.button_press_reply,
                        SloppyFocus.pre_button_press_reply)

class ClickToSloppyFocus(ClickToFocus, SloppyFocus):
    "Same as SloppyFocus, but put the clicked window on top."
    def init_focus(self, wm):
        # same beginning as SloppyFocus
        SloppyFocus.init_focus(self, wm)
        # (nearly) same end as ClickToFocus
        ClickToFocus.init_focus_aux(self, wm)

    @staticmethod
    @managed_client
    def pre_button_press_reply(wm, event, client):
        # pre hook too, for the same reason as ClickToFocus
        if event.window == client.window:
            wm.display.allow_events(X.ReplayPointer, X.CurrentTime)
            client.upper()

        SloppyFocus.pre_button_press_reply(wm, event)

        # make sure the client is focused
        client.take_focus()

    @staticmethod
    @managed_client
    def map_request_reply(wm, event, client):
        # if a client has been mapped, do nothing as in SloppyFocus
        pass

    def destroy_focus(self, wm):
        SloppyFocus.destroy_focus(self, wm)
        ClickToFocus.destroy_focus_aux(self, wm)
